/**
 * SPDX-FileCopyrightText: 2020-2023 Osimis S.A., 2024-2026 Orthanc Team SRL, 2021-2026 Sebastien Jodogne, ICTEAM UCLouvain
 * SPDX-License-Identifier: AGPL-3.0-or-later
 */

/**
 * Python plugin for Orthanc
 * Copyright (C) 2020-2023 Osimis S.A., Belgium
 * Copyright (C) 2024-2026 Orthanc Team SRL, Belgium
 * Copyright (C) 2021-2026 Sebastien Jodogne, ICTEAM UCLouvain, Belgium
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 **/


#pragma once

#include "PythonObject.h"

#include <Compatibility.h>

#include <memory>

class PythonModule;

class PythonFunction : public boost::noncopyable
{
  friend class PythonLock;
  
private:
  PythonLock&                    lock_;
  std::unique_ptr<PythonObject>  func_;

  PythonObject* CallUnchecked(PyObject* args);

public:
  PythonFunction(PythonLock& lock,
                 const PythonModule& module,
                 const std::string& name);

  bool IsValid() const
  {
    return func_.get() != NULL;
  }

  PythonObject* Call();

  PythonObject* Call(const PythonObject& args);

  void CallVoid()
  {
    std::unique_ptr<PythonObject> result(Call());
  }
};
