/**
 * SPDX-FileCopyrightText: 2020-2023 Osimis S.A., 2024-2026 Orthanc Team SRL, 2021-2026 Sebastien Jodogne, ICTEAM UCLouvain
 * SPDX-License-Identifier: AGPL-3.0-or-later
 */

/**
 * Python plugin for Orthanc
 * Copyright (C) 2020-2023 Osimis S.A., Belgium
 * Copyright (C) 2024-2026 Orthanc Team SRL, Belgium
 * Copyright (C) 2021-2026 Sebastien Jodogne, ICTEAM UCLouvain, Belgium
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU Affero General Public License
 * as published by the Free Software Foundation, either version 3 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Affero General Public License for more details.
 * 
 * You should have received a copy of the GNU Affero General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 **/


// WARNING: Auto-generated file. Do not modify it by hand.


#include <PythonHeaderWrapper.h>  // Must be before "sdk.h"
#include "./sdk.h"

#include <PythonLock.h>
#include <PythonThreadsAllower.h>

#include <OrthancPluginCppWrapper.h>

{{#enumerations}}
#include "./sdk_{{name}}.impl.h"
{{/enumerations}}

{{#classes}}
#include "./sdk_{{class_name}}.impl.h"
{{/classes}}

#include "./sdk_GlobalFunctions.impl.h"

{{#classes}}
#include "./sdk_{{class_name}}.methods.h"
{{/classes}}

void RegisterOrthancSdk(PyObject* module)
{
{{#enumerations}}
  Register{{name}}Enumeration(module);
{{/enumerations}}

{{#classes}}
  Register{{class_name}}Class(module);
{{/classes}}
}
